@@ -12,32 +12,37 @@ from .data import energy_path, energy_path_from_p_state
 from .util import set_output_folder
 from .chain_io import chain_write_between
 from datetime import datetime
+import uuid
 import json
 
 class GNEB_Node(NodeMixin):
     """A class that represents a GNEB calculation on a single chain. Can spawn children if cutting of the chain becomes necessary."""
 
-    def __init__(self, name, input_file, output_folder, initial_chain_file=None, gneb_workflow_log_file=None, parent=None, children=None):
+    def __init__(self, name, input_file, output_folder, initial_chain_file=None, gneb_workflow_log_file=None, parent=None, children=None, scratch_folder=None):
         from spirit import simulation, io
 
         """Constructor."""
         self.chain_file : str             = ""
+        self.chain_file_post_run : str    = ""
         self.initial_chain_file : str     = ""
         self.input_file : str             = ""
         self.gneb_workflow_log_file : str = ""
+        self.uuid                = uuid.uuid1()
         self.current_energy_path = object()
         self.n_iterations_check  = 5000
         self.n_checks_save       = 3
         self.total_iterations    = 0
-        self.target_noi  = 10
+        self.target_noi  = -1
         self.noi         = -1
-        self.convergence = 1e-4
+        self.convergence = 1e-7
         self.path_shortening_constant = -1
         self.max_total_iterations     = -1
         self.output_folder = ""
         self.output_tag    = ""
-        self.allow_split = True
-        self.moving_endpoints       = False
+        self.allow_split   = False
+        self.max_depth     = 1
+        self.moving_endpoints       = None
+        self.translating_endpoints  = None
         self.delta_Rx_left          = 1.0
         self.delta_Rx_right         = 1.0
         self.state_prepare_callback = None
@@ -45,6 +50,7 @@ class GNEB_Node(NodeMixin):
         self.exit_callback          = None
         self.before_gneb_callback   = None
         self.before_llg_callback    = None
+        self.after_llg_callback     = None
         self.intermediate_minima = []
         self.child_indices       = []
         self.image_types         = []
@@ -52,19 +58,30 @@ class GNEB_Node(NodeMixin):
         self.solver_llg  = simulation.SOLVER_LBFGS_OSO
         self.solver_gneb = simulation.SOLVER_VP_OSO
         self.chain_write_fileformat = io.FILEFORMAT_OVF_BIN
-
-        # private fields
-        self._ci                 = False
-        self._converged          = False
-        self._log                = True # flag to disable/enable logging
+        self.input_file = input_file
+        self.name       = name
+        self.parent     = parent
+        self._log       = True # flag to disable/enable logging
+        self._converged = False
 
         # Create output folder
         if not os.path.exists(output_folder):
             os.makedirs(output_folder)
         self.output_folder = output_folder
 
+        # If no scratch folder is specified, we use the output folder
+        # else we create a temporary folder in the scratch partition
+        if scratch_folder is None:
+            self.scratch_folder = self.output_folder
+        else:
+            self.scratch_folder = os.path.join(scratch_folder, self.name + "#" + str(self.uuid))
+            os.makedirs(self.scratch_folder)
+
         # The current chain is always saved here
-        self.chain_file = output_folder + "/chain.ovf"
+        self.chain_file = self.scratch_folder + "/chain.ovf"
+
+        # After a run, we copy the chain to this file
+        self.chain_file_post_run = self.output_folder + "/chain.ovf"
 
         # If an initial chain is specified we copy it to the output folder
         if(initial_chain_file):
@@ -73,10 +90,6 @@ class GNEB_Node(NodeMixin):
             self.initial_chain_file = initial_chain_file
             shutil.copyfile(initial_chain_file, self.chain_file)
 
-        self.input_file = input_file
-        self.name       = name
-        self.parent     = parent
-
         # If no log file has been specified we put one in the output folder
         if not gneb_workflow_log_file:
             self.gneb_workflow_log_file = self.output_folder + "/workflow_log.txt"
@@ -96,6 +109,9 @@ class GNEB_Node(NodeMixin):
                 creation_msg += "{} ".format(c.name)
         self.log(creation_msg)
 
+        if(initial_chain_file):
+            self.log("Initial chain file: {}".format(initial_chain_file))
+
     def setup_plot_callbacks(self):
         """Sets up callbacks such that the path is plotted."""
         from spirit import simulation, chain
@@ -140,6 +156,13 @@ class GNEB_Node(NodeMixin):
             plt.close()
         self.gneb_step_callback = step_cb
 
+        def after_llg_cb(gnw, p_state):
+            plotting.plot_energy_path(gnw.current_energy_path, plt.gca())
+            mark_climbing_image(p_state, gnw, plt.gca())
+            plt.savefig(gnw.output_folder + "/path_after_llg.png".format(gnw.total_iterations))
+            plt.close()
+        self.after_llg_callback = after_llg_cb
+
         def exit_cb(gnw, p_state):
             plotting.plot_energy_path(gnw.current_energy_path, plt.gca())
             mark_climbing_image(p_state, gnw, plt.gca())
@@ -156,10 +179,15 @@ class GNEB_Node(NodeMixin):
             log_file = new_output_folder + "/workflow_log.txt"
 
         self.output_folder          = new_output_folder
-        self.chain_file             = self.output_folder + "/chain.ovf"
+
+        # Only change chain file, if it was previously in the output folder
+        if self.chain_file == self.chain_file_post_run:
+            self.chain_file = self.output_folder + "/chain.ovf"
+        self.chain_file_post_run = self.output_folder + "/chain.ovf"
+
         self.gneb_workflow_log_file = log_file
         if os.path.exists(self.initial_chain_file):
-            self.initial_chain_file     = self.output_folder + "/root_initial_chain.ovf"
+            self.initial_chain_file  = self.output_folder + "/root_initial_chain.ovf"
 
         self.log("Changed output folder to {}".format(new_output_folder))
 
@@ -218,7 +246,9 @@ class GNEB_Node(NodeMixin):
 
         node_dict = dict(
             name                     = str(self.name),
+            uuid                     = str(self.uuid),
             chain_file               = str(self.chain_file),
+            chain_file_post_run      = str(self.chain_file_post_run),
             initial_chain_file       = str(self.initial_chain_file),
             input_file               = str(self.input_file),
             gneb_workflow_log_file   = str(self.gneb_workflow_log_file),
@@ -229,9 +259,11 @@ class GNEB_Node(NodeMixin):
             convergence              = float(self.convergence),
             max_total_iterations     = int(self.max_total_iterations),
             output_folder            = str(self.output_folder),
+            scratch_folder           = str(self.scratch_folder),
             output_tag               = str(self.output_tag),
             child_indices            = [(int(c1), int(c2)) for c1, c2 in self.child_indices],
             allow_split              = bool(self.allow_split),
+            max_depth                = int(self.max_depth),
             path_shortening_constant = float(self.path_shortening_constant),
             moving_endpoints         = bool(self.moving_endpoints),
             delta_Rx_left            = float(self.delta_Rx_left),
@@ -248,6 +280,9 @@ class GNEB_Node(NodeMixin):
         for c in self.children:
             c.to_json()
 
+    def history_to_file(self, path):
+        np.savetxt(path, self.history, header="iteration, max. torque" )
+
     @staticmethod
     def from_json(json_file, parent=None, children=None):
 
@@ -270,6 +305,7 @@ class GNEB_Node(NodeMixin):
         result.initial_chain_file       = data.get("initial_chain_file", result.initial_chain_file)
         result.output_tag               = data.get("output_tag", result.output_tag)
         result.allow_split              = data.get("allow_split", result.allow_split)
+        result.max_depth                = data.get("max_depth", result.max_depth)
         result.child_indices            = data.get("child_indices", result.child_indices)
         result.moving_endpoints         = data.get("moving_endpoints", result.moving_endpoints)
         result.image_types              = data.get("image_types", result.image_types)
@@ -306,15 +342,18 @@ class GNEB_Node(NodeMixin):
 
     def update_energy_path(self, p_state=None):
         """Updates the current energy path. If p_state is given we just use that, otherwise we have to construct it first"""
+        self._log = False
         if p_state:
             self.current_energy_path = energy_path_from_p_state(p_state)
             self.noi = self.current_energy_path.noi()
         else:
-            from spirit import state, io, chain
+            from spirit import state, chain, simulation
             with state.State(self.input_file) as p_state:
                 chain.update_data(p_state)
                 self._prepare_state(p_state)
+                simulation.start(p_state, simulation.METHOD_GNEB, self.solver_gneb, n_iterations=1) # One iteration of GNEB to get interpolated quantities
                 self.update_energy_path(p_state)
+        self._log = True
 
     def check_for_minima(self):
         """
@@ -335,13 +374,16 @@ class GNEB_Node(NodeMixin):
         self.log("Writing chain to {}".format(self.chain_file))
         io.chain_write(p_state, self.chain_file, fileformat=self.chain_write_fileformat)
 
-    def backup_chain(self, path):
+    def backup_chain(self, path, p_state=None):
         """Saves the chain to a file"""
-        from spirit import state, io
-        with state.State(self.input_file) as p_state:
-            self._prepare_state(p_state)
+        if not p_state is None:
             self.log("Writing chain to {}".format(path))
             io.chain_write(p_state, path, fileformat=self.chain_write_fileformat)
+        else:
+            from spirit import state, io
+            with state.State(self.input_file) as p_state:
+                self._prepare_state(p_state)
+                self.backup_chain(path)
 
     def add_child(self, i1, i2, p_state=None):
 
@@ -351,22 +393,30 @@ class GNEB_Node(NodeMixin):
             child_name          = self.name + "_{}".format(len(self.children))
             child_input_file    = self.input_file
             child_output_folder = self.output_folder + "/{}".format(len(self.children))
-            self.children      += (GNEB_Node(name = child_name, input_file = child_input_file, output_folder = child_output_folder, gneb_workflow_log_file=self.gneb_workflow_log_file, parent = self), )
+
+            if self.scratch_folder == self.output_folder:
+                child_scratch_folder = None
+            else:
+                child_scratch_folder = self.scratch_folder
+
+            self.children      += (GNEB_Node(name = child_name, input_file = child_input_file, output_folder = child_output_folder, scratch_folder = child_scratch_folder, gneb_workflow_log_file = self.gneb_workflow_log_file, parent = self), )
             self.children[-1].current_energy_path = self.current_energy_path.split(i1, i2+1)
             # Copy the other attributes
-            self.children[-1].target_noi             = self.target_noi
-            self.children[-1].convergence            = self.convergence
+            self.children[-1].target_noi               = self.target_noi
+            self.children[-1].convergence              = self.convergence
             self.children[-1].path_shortening_constant = self.path_shortening_constant
-            self.children[-1].max_total_iterations   = self.max_total_iterations
-            self.children[-1].state_prepare_callback = self.state_prepare_callback
-            self.children[-1].gneb_step_callback     = self.gneb_step_callback
-            self.children[-1].exit_callback          = self.exit_callback
-            self.children[-1].before_gneb_callback   = self.before_gneb_callback
-            self.children[-1].before_llg_callback    = self.before_llg_callback
-            self.children[-1].n_iterations_check     = self.n_iterations_check
-            self.children[-1].n_checks_save          = self.n_checks_save
-            self.children[-1].allow_split            = self.allow_split
-            self.children[-1].chain_write_fileformat = self.chain_write_fileformat
+            self.children[-1].max_total_iterations     = self.max_total_iterations
+            self.children[-1].state_prepare_callback   = self.state_prepare_callback
+            self.children[-1].gneb_step_callback       = self.gneb_step_callback
+            self.children[-1].exit_callback            = self.exit_callback
+            self.children[-1].before_gneb_callback     = self.before_gneb_callback
+            self.children[-1].before_llg_callback      = self.before_llg_callback
+            self.children[-1].after_llg_callback       = self.after_llg_callback
+            self.children[-1].n_iterations_check       = self.n_iterations_check
+            self.children[-1].n_checks_save            = self.n_checks_save
+            self.children[-1].allow_split              = self.allow_split
+            self.children[-1].max_depth                = self.max_depth
+            self.children[-1].chain_write_fileformat   = self.chain_write_fileformat
 
             self.child_indices.append([i1, i2])
             # Write the chain file
@@ -382,25 +432,53 @@ class GNEB_Node(NodeMixin):
 
         return self.children[-1] # Return a reference to the child that has just been added
 
+    def exit(self, p_state):
+        """Call this when p_state gets deleted"""
+        self.save_chain(p_state)
+
+        if(self.exit_callback):
+            self.exit_callback(self, p_state)
+
+        if self.chain_file != self.chain_file_post_run:
+            self.log("Copying chain_file {} to {}".format(self.chain_file, self.chain_file_post_run))
+            shutil.copyfile(self.chain_file, self.chain_file_post_run)
+
+    def relax_intermediate_minima(self, p_state=None):
+        from spirit import state, chain, simulation
+        if not p_state is None:
+            self.log("Relaxing intermediate minima")
+            for idx_minimum in self.intermediate_minima:
+                simulation.start(p_state, simulation.METHOD_LLG, self.solver_llg, idx_image = idx_minimum)
+            chain.update_data(p_state)
+            self.update_energy_path(p_state)
+            if self.after_llg_callback:
+                self.after_llg_callback(self, p_state)
+        else:
+            with state.State(self.input_file) as p_state:
+                self._log = False
+                self._prepare_state(p_state)
+                chain.update_data(p_state)
+                self.check_for_minima()
+                self._log = True
+                self.relax_intermediate_minima(p_state)
+                self.update_energy_path(p_state)
+                self.exit(p_state)
+
     def spawn_children(self, p_state):
         """If intermediate minima are present, relaxes them and creates child nodes"""
-
         from spirit import chain, simulation
 
-        if not self.allow_split:
+        if not self.allow_split or self.depth >= self.max_depth:
             return
 
         if len(self.intermediate_minima) == 0:
             return
 
-        if self._converged:
-            return
+        if self.before_llg_callback:
+            self.before_llg_callback(self, p_state)
 
         self.log("Spawning children - splitting chain")
-
-        self.log("Relaxing intermediate minima")
-        for idx_minimum in self.intermediate_minima:
-            simulation.start(p_state, simulation.METHOD_LLG, self.solver_llg, idx_image = idx_minimum)
+        self.relax_intermediate_minima(p_state)
 
         # Instantiate the GNEB nodes
         noi = chain.get_noi(p_state)
@@ -421,12 +499,6 @@ class GNEB_Node(NodeMixin):
 
         idx_children_run_order = list(range(len(self.children)))
 
-        try:
-            idx_children_run_order.sort(key = lambda i : -self.children[i].current_energy_path.barrier())
-        except Exception as e:
-            self.log("Could not sort children run order by energy barrier. Executing in order.")
-            idx_children_run_order = list(range(len(self.children)))
-
         for i in idx_children_run_order:
             self.children[i].run()
 
@@ -491,7 +563,10 @@ class GNEB_Node(NodeMixin):
 
         if self.moving_endpoints is not None:
             parameters.gneb.set_moving_endpoints(p_state, self.moving_endpoints)
-            parameters.gneb.set_equilibrium_delta_Rx(p_state, self.delta_Rx_left, self.delta_Rx_right)
+            if self.translating_endpoints is not None:
+                parameters.gneb.set_translating_endpoints(p_state, self.translating_endpoints)
+            if self.delta_Rx_left is not None and self.delta_Rx_right is not None:
+                parameters.gneb.set_equilibrium_delta_Rx(p_state, self.delta_Rx_left, self.delta_Rx_right)
 
         # The state prepare callback can be used to change the state before execution of any other commands
         # One could e.g. use the hamiltonian API to change interaction parameters instead of relying only on the input file
@@ -519,7 +594,7 @@ class GNEB_Node(NodeMixin):
         condition_iterations = self.total_iterations < self.max_total_iterations or self.max_total_iterations < 0
         condition_convergence = not self._converged
 
-        if self.allow_split: # Only care about minima if splitting is allowed
+        if self.allow_split and self.depth < self.max_depth: # Only care about minima if splitting is allowed
             result = condition_minima and condition_iterations and condition_convergence
         else:
             result = condition_iterations and condition_convergence
@@ -592,24 +667,17 @@ class GNEB_Node(NodeMixin):
 
                 except KeyboardInterrupt as e:
                     self.log("Interrupt during run loop")
-                    self.save_chain(p_state)
-                    if(self.exit_callback):
-                        self.exit_callback(self, p_state)
+                    self.exit(p_state)
                     raise e
 
-                if self.before_llg_callback:
-                    self.before_llg_callback(self, p_state)
-
                 self.update_energy_path(p_state)
 
-                self.spawn_children(p_state)
-
-                self.save_chain(p_state)
-
-                if(self.exit_callback):
-                    self.exit_callback(self, p_state)
+                if not self._converged:
+                    self.spawn_children(p_state)
 
+                self.exit(p_state)
                 # p_state gets deleted here, it does not have to persist to run the child nodes
+
             self.run_children()
             self.log("Finished!")
 
@@ -623,10 +691,6 @@ class GNEB_Node(NodeMixin):
 
         self.log("Preparing for moving endpoints")
 
-        # chain_backup_path = os.path.join(self.output_folder, "chain_backup.ovf")
-        # self.log("    Making backup copy of chain in {}".format(chain_backup_path))
-        # shutil.copy(self.chain_file, chain_backup_path)
-
         self.target_noi       = 3
         self.moving_endpoints = True
         self.image_types      = [[1, parameters.gneb.IMAGE_CLIMBING]]
@@ -641,7 +705,7 @@ class GNEB_Node(NodeMixin):
                 E = chain.get_energy(p_state)
                 idx_mid = np.argmax(E)
 
-            self.log("    idx_mid = {}".format(idx_mid))
+            self.log("idx_mid = {}".format(idx_mid))
 
             if(idx_mid >= 1 and idx_mid < noi-1):
                 for i in range(idx_mid-1):
@@ -652,88 +716,38 @@ class GNEB_Node(NodeMixin):
             self.update_energy_path(p_state)
             self.save_chain(p_state)
 
-    def clamp_and_refine(self, convergence=None, max_total_iterations=None, idx_max_list=None, apply_ci=True, target_noi=5):
-        """One step of clamp and refine algorithm"""
+    def prepare_dimer(self, idx_left, idx_right=None):
+        from spirit import chain, io, state, parameters
 
-        if target_noi % 2 == 0:
-            raise Exception("target_noi must be uneven!")
+        if idx_right is None:
+            idx_right = idx_left + 1
 
-        if not convergence:
-            convergence = self.convergence
+        self.log("Preparing for dimer endpoints")
 
-        if not max_total_iterations:
-            max_total_iterations = self.max_total_iterations
+        self.target_noi            = 2
+        self.moving_endpoints      = True
+        self.translating_endpoints = True
+        self.image_types           = []
 
-        try:
-            from spirit import state
-            if(len(self.children) != 0): # If not a leaf node call recursively on children
-                for c in self.children:
-                    c.clamp_and_refine(convergence, max_total_iterations, idx_max_list, apply_ci, target_noi)
-                return
+        with state.State(self.input_file) as p_state:
+            self._prepare_state(p_state)
+            self.update_energy_path(p_state)
 
-            self.update_energy_path()
+            noi = chain.get_noi(p_state)
 
-            # To get a list of the "interesting" images we compute the second derivative of the energy wrt to Rx
-            # We initialize these lists with a dummy element so the idx counting is not offset by one
-            second_deriv    = [0] # second derivative
-            first_deriv_fw  = [0] # first derivative forward
-            first_deriv_bw  = [0] # first derivative backward
+            self.log("idx_left = {}, idx_right = {}".format(idx_left, idx_right))
 
-            E  = self.current_energy_path.total_energy
-            Rx = self.current_energy_path.reaction_coordinate
+            # Delete all images to the right of idx right
+            for i in range(noi - idx_right - 1):
+                chain.pop_back(p_state)
 
-            for idx in range(1, self.current_energy_path.noi()-1): # idx=0 and idx=noi-1 excluded!
-                grad_forward = (E[idx+1] - E[idx]) / (Rx[idx+1] - Rx[idx])
-                grad_backward = (E[idx] - E[idx-1]) / (Rx[idx] - Rx[idx-1])
-                second = 2 * (grad_forward - grad_backward) / ( Rx[idx+1] - Rx[idx-1] )
-                first_deriv_fw.append(grad_forward)
-                first_deriv_bw.append(grad_backward)
-                second_deriv.append(second)
+            # Delete all images to the left of idx_left
+            for i in range(idx_left):
+                chain.delete_image(p_state, idx_image=0)
 
-            # If not specified, build the idx_max list
-            if not idx_max_list:
-                idx_max_list = []
-                idx = 1
-                while idx<self.current_energy_path.noi()-1:
-                    if first_deriv_fw[idx]<0 and first_deriv_bw[idx] > 0.01 * first_deriv_fw[idx]:
-                        idx_max_list.append(idx)
-                        idx += 1 # If we add something to idx_max list we increment the idx so that we do not also add the point right after that
-                    idx += 1
+            # Delete images between idx_left and idx_right
+            for i in range(idx_right - idx_left - 1):
+                chain.delete_image(p_state, idx_image=1)
 
-            self.log("Clamp and refine! idx_list = {}".format(idx_max_list))
-
-            for idx_max in idx_max_list:
-                if(idx_max == 0 or idx_max == self.noi - 1):
-                    self.log("Cannot clamp and refine, since idx_max (= {}) is either 0 or noi-1".format(idx_max))
-                    return
-
-                with state.State(self.input_file) as p_state:
-                    self._prepare_state(p_state)
-                    self.add_child(idx_max-1, idx_max+1, p_state)
-
-                    # Attributes, that we dont copy
-                    self.children[-1].allow_split          = False  # Dont allow splitting for clamp and refine
-                    self.children[-1].path_shortening_constant = -1 # No path shortening for clamp and refine
-                    self.children[-1].target_noi           = target_noi
-                    self.children[-1].convergence          = convergence
-                    self.children[-1].max_total_iterations = max_total_iterations
-
-                    if apply_ci and not self._ci:
-                        def before_gneb_cb(gnw, p_state):
-                            from spirit import parameters
-                            self.before_gneb_callback(gnw, p_state)
-                            gnw.log("Setting image type")
-                            parameters.gneb.set_climbing_falling(p_state, parameters.gneb.IMAGE_CLIMBING, idx_image=int((target_noi-1)/2))
-                        self.children[-1]._ci = True
-                    else:
-                        def before_gneb_cb(gnw, p_state):
-                            self.before_gneb_callback(gnw, p_state)
-
-                    self.children[-1].before_gneb_callback = before_gneb_cb
-
-            self.run_children()
-
-        except Exception as e:
-            self.log("Exception during 'clamp_and_refine': {}".format(str(e))) # Log the exception and re-raise
-            self.log(traceback.format_exc())
-            raise e
\ No newline at end of file
+            self.update_energy_path(p_state)
+            self.save_chain(p_state)
\ No newline at end of file
